import { Service } from "../service"
import * as mongodb from "mongodb"
import * as Leaf from "leaf-ts"
export const Errors = Leaf.ErrorDoc.create()
    .define("NotAvailable")
    .generate() as Leaf.ErrorsOf<{
        NotAvailable: {}
    }>
export class MemoryLockService extends Service {
    name: "MemoryLockService"
    private locks: {
        [key: string]: any
    } = {}
    constructor() {
        super()
    }
    public lock(option: {
        resource: string
    }, callback: (err: Error, lock?: string) => void) {
        if (!this.locks[option.resource]) {
            let lock = this.locks[option.resource] = Math.random().toString().slice(4)
            callback(null, lock)
        } else {
            callback(new Errors.NotAvailable())
        }
    }
    private counter = 0
    private GC_COUNT
    public release(resource: string, lock: string, callback: Function = () => { }) {
        if (this.locks[resource] !== lock) {
            callback()
            return false
        }
        this.locks[resource] = null
        this.counter++
        if (this.counter >= this.GC_COUNT) {
            this.gc()
        }
        callback()
        return true
    }
    // defer callback
    public defer<T extends Function>(resource: string, lock: string, callback: T): T {
        let fn = (...args: any[]) => {
            this.release(resource, lock, () => {
                callback(...args)
            })
        }
        return fn as any as T
    }
    private gc() {
        let locks = {}
        // Here we don't delege empty field, we just build a new locks.
        // There are usually very few living locks
        for (let prop in this.locks) {
            let v = this.locks[prop]
            if (v) {
                locks[prop] = v
            }
        }
        this.locks = locks
    }
}
